/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.dao.ContactInformationDAO;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.dao.impl.abstr.AbstractDAOImpl;
import dk.sds.ecpr2.domain.ContactInformationVO;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.stereotype.Repository;

import java.util.Date;
import java.util.List;

@Repository
public class ContactInformationDAOImpl extends AbstractDAOImpl implements ContactInformationDAO {

    public static final String TABLE_NAME = "ContactInformations";

    @Override
    public ContactInformationVO getContactInformation(String cprNo, Date date) throws DAOException {
        final List<ContactInformationVO> contactInfoList = getByCprAndDate(TABLE_NAME, cprNo, date, new ContactInformationAndDataCardRowMapper());
        throwExceptionIfMoreThanOne(contactInfoList);
        return firstElementOrNull(contactInfoList);
    }

    @Override
    public ContactInformationVO getContactInformationWithoutDataCard(String cprNo, Date date) throws DAOException {
        final List<ContactInformationVO> contactInfoList;
        try {
            final String sql = "SELECT ci.* FROM ContactInformations ci, PersonalDataCards pdc "
                    + "WHERE pdc.PersonIdentifier = :CPR "
                    + "AND ci.PersonalDataCardPID = pdc.PersonalDataCardPID "
                    + "AND ci.ValidFrom <= :Date AND ci.ValidTo > :Date "
                    + "ORDER BY Version LIMIT 1";
            final MapSqlParameterSource parms = new MapSqlParameterSource()
                    .addValue("CPR", cprNo)
                    .addValue("Date", date);
            contactInfoList = executeQuery(new ContactInformationRowMapper(), sql, parms);

            throwExceptionIfMoreThanOne(contactInfoList);

        } catch (Exception e) {
            throw new DAOException("An error occurred loading contact information by time", e);
        }

        return firstElementOrNull(contactInfoList);
    }

    @Override
    public List<ContactInformationVO> getAllVersionsOfContactInformationWithoutDataCard(String cprNo) throws DAOException {
        final List<ContactInformationVO> contactInfoList;
        try {
            final String sql = "SELECT ci.* FROM ContactInformations ci, PersonalDataCards pdc "
                    + "WHERE pdc.PersonIdentifier = :CPR "
                    + "AND ci.PersonalDataCardPID = pdc.PersonalDataCardPID "
                    + "ORDER BY Version";
            final MapSqlParameterSource parameters = new MapSqlParameterSource()
                .addValue("CPR", cprNo);
            contactInfoList = executeQuery(new ContactInformationRowMapper(), sql, parameters);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading all versions of contact information", e);
        }
        return contactInfoList;
    }

    @Override
    public void updateValidToTimestamp(long dataCardId, int version, Date timestamp) throws DAOException {
        try {
            final String sql = "UPDATE ContactInformations SET ValidTo = :ValidTo WHERE PersonalDataCardPID = :PersonalDataCardPID AND Version = :Version";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("ValidTo", timestamp);
            parameters.addValue("Version", version);
            parameters.addValue("PersonalDataCardPID", dataCardId);
            executeUpdate(sql, parameters);
        } catch (Exception e) {
            throw new DAOException("An error updating valid to timestamp", e);
        }
    }

    @Override
    public void createContactInformation(ContactInformationVO contactInformation) throws DAOException {
        try {
            final String sql = "INSERT INTO ContactInformations (PersonalDataCardPID,Version,ValidFrom,ValidTo,MobilePhone,HomePhone,WorkPhone) " +
                    "VALUES (:PersonalDataCardPID,:Version,:ValidFrom,:ValidTo,:MobilePhone,:HomePhone,:WorkPhone)";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();
            parameters.addValue("PersonalDataCardPID", contactInformation.getPersonalDataCardPID());
            parameters.addValue("Version", contactInformation.getVersion());
            parameters.addValue("ValidFrom", contactInformation.getValidFrom());
            parameters.addValue("ValidTo", contactInformation.getValidTo());
            parameters.addValue("MobilePhone", contactInformation.getMobilePhone());
            parameters.addValue("HomePhone", contactInformation.getHomePhone());
            parameters.addValue("WorkPhone", contactInformation.getWorkPhone());
            executeUpdate(sql, parameters);
        } catch (Exception e) {
            throw new DAOException("An error occurred creating contactInformations", e);
        }
    }

    private void throwExceptionIfMoreThanOne(List<ContactInformationVO> contactInfoList) throws DAOException {
        final int count = contactInfoList.size();
        if (count > 1) {
            throw new DAOException("There should only be one active contact information at a single point in time. Found: " + count);
        }
    }

    private ContactInformationVO firstElementOrNull(List<ContactInformationVO> contactInfoList) {
        return contactInfoList.size() > 0 ? contactInfoList.get(0) : null;
    }

}
