/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.IdAndVersionVO;
import dk.sds.ecpr2.domain.abstr.VersionedAbstractVO;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.actuate.metrics.CounterService;
import org.springframework.boot.actuate.metrics.GaugeService;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.util.*;

/**
 * Base class for DAO implementations
 */
@Log4j2
public class AbstractDAOImpl {
    protected static final String ACTOR_FIELDS = "a.ServiceActorPID, a.PersonIdentifier as ActorPersonIdentifier, a.PersonIdentifierSource as ActorPersonIdentifierSource, a.FirstName, a.LastName, a.AuthorisationNumber, a.EducationCode, a.SpecialityCode, a.SpecialityCodeSource, a.OrganizationId, a.OrganizationIdSource, a.OrganizationName, a.HashValue";
    protected static final String TS_AND_ACTOR_FIELDS = "p.ModifiedDateTime, " + ACTOR_FIELDS;

    @Autowired
    private DataSource dataSource;

    @Autowired
    private GaugeService gaugeService;

    @Autowired
    private CounterService counterService;

    protected NamedParameterJdbcTemplate namedParamJdbcTemplate;

    final List<String> tableWhiteList = Arrays.asList("Relatives", "TempAddresses", "ContactInformations", "Languages");


    @PostConstruct
    private void init() {
        namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
    }

    protected List executeQuery(RowMapper rowMapper, String sql, MapSqlParameterSource parms) {
        final long startTime = System.nanoTime();
        final List list = namedParamJdbcTemplate.query(sql, parms, rowMapper);
        recordMetrics(sql, startTime);
        return list;
    }

    protected void executeUpdate(String sql, MapSqlParameterSource parameters) {
        final long startTime = System.nanoTime();
        namedParamJdbcTemplate.update(sql, parameters);
        recordMetrics(sql, startTime);
    }

    protected int[] executeBatchUpdate(String sql, SqlParameterSource[] batch) {
        final long startTime = System.nanoTime();
        final int[] result =  namedParamJdbcTemplate.batchUpdate(sql, batch);
        recordMetrics(sql, startTime);
        return result;
    }

    protected <T extends VersionedAbstractVO>  List<T> getByCprAndDate(String tableName, String cprnr, Date date, RowMapper rowMapper) throws DAOException {
        try {
           checkLegalTableName(tableName);
            final String sql = "SELECT t.*," + TS_AND_ACTOR_FIELDS + " FROM PersonalDataCards p, ServiceActors a, " + tableName + " t" +
                    " WHERE p.PersonIdentifier = :cprnr" +
                    " AND t.PersonalDataCardPID = p.PersonalDataCardPID" +
                    " AND t.ValidFrom <= :date " +
                    " AND t.ValidTo > :date " +
                    " AND p.ServiceActorPID = a.ServiceActorPID";
            final MapSqlParameterSource parms = new MapSqlParameterSource()
                    .addValue("cprnr", cprnr)
                    .addValue("date", date);
            final List<T> list = executeQuery(rowMapper, sql, parms);
            return list;
        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + tableName + " by CPR", e);
        }
    }

    protected void updateValidToTimestamp(String tableName, String idColumnName, List<IdAndVersionVO> idAndVersionList, Date timestamp) throws DAOException {
        if (idAndVersionList.isEmpty()) {
            return;
        }

        try {
            final List<String> columnWhiteList = Arrays.asList("Uuid", "Id");
            if ( ! (tableWhiteList.contains(tableName) && columnWhiteList.contains(idColumnName)) ) {
                throw new SecurityException("Potential SQL injection attack");
            }

            String sql = "UPDATE " + tableName + " SET ValidTo = :ValidTo WHERE ";
            final MapSqlParameterSource parameters = new MapSqlParameterSource();

            final List<String> conditions = new LinkedList<>();
            for (IdAndVersionVO idAndVersion: idAndVersionList) {
                conditions.add(idColumnName + " = '" + idAndVersion.getId() + "' AND Version = " + idAndVersion.getVersion());
                parameters.addValue("ValidTo", timestamp);
                parameters.addValue("Ids", idAndVersion);
            }

            sql += String.join(" OR ", conditions);
            executeUpdate(sql, parameters);

        } catch (Exception e) {
            throw new DAOException("An error occurred updating valid to timestamps", e);
        }
    }


    protected <T extends VersionedAbstractVO> List<T> getByUuidsAndDateWithoutDataCard(String tableName, RowMapper rowMapper, List<String> uuids, Date date) throws DAOException {
        final List<T> list;
        if (date == null) {
            throw new DAOException("The specified timestamp was null.");
        }
        if (uuids == null) {
            throw new DAOException("The specified UUID list was null.");
        }
        if (uuids.size() == 0) {
            return new ArrayList<T>();
        }

        try {
            checkLegalTableName(tableName);
            final String sql = "SELECT * FROM " + tableName + " WHERE  ValidFrom <= :Date AND ValidTo > :Date AND Uuid IN (:Uuids)";
            final MapSqlParameterSource parms = new MapSqlParameterSource()
                    .addValue("Date", date)
                    .addValue("Uuids", uuids);
            list = executeQuery(rowMapper, sql, parms);

        } catch (Exception e) {
            throw new DAOException("An error occurred loading " + tableName + " by uuid and time", e);
        }
        return list;
    }


    protected <T extends VersionedAbstractVO> List<T> getAllVersionsByUuidsWithoutDataCard(String tableName, RowMapper rowMapper, List<String> uuids) throws DAOException {
        final List<T> list;

        if (uuids == null) {
            throw new DAOException("The specified UUID list was null.");
        }
        if (uuids.size() == 0) {
            return new ArrayList<T>();
        }

        try {
            checkLegalTableName(tableName);
            final String sql = "SELECT * FROM " + tableName + " WHERE Uuid IN (:Uuids) ORDER BY Version";
            final MapSqlParameterSource parms = new MapSqlParameterSource()
                    .addValue("Uuids", uuids);
            list = executeQuery(rowMapper, sql, parms);
        } catch (Exception e) {
            throw new DAOException("An error occurred loading all versions of " + tableName + " by uuid", e);
        }
        return list;
    }


    private void checkLegalTableName(String tableName) {
        if (!(tableWhiteList.contains(tableName))) {
            throw new SecurityException("Potential SQL injection attack. Unrecognized table name: " + tableName);
        }
    }

    private void recordMetrics(String sql, long startTime) {
        final String metricName = sql.substring(0, 50);
        final double durationMillis = (System.nanoTime() - startTime)/1000000;
        gaugeService.submit(metricName, durationMillis);
        counterService.increment(metricName);
    }



}
